(** * Completeness of [LJT] w.r.t. Kripke Semantics *)

Set Implicit Arguments.

Require Import decidable_IN.
Require Import associations.
Require Import sublist.
Require Import language_syntax.
Require Import variable_sets.
Require Import substitution.
Require Import fresh_parameters.
Require Import relocation.  
Require Import LJT_and_weakening.
Require Import kripke_semantics.
Require Import soundness.

(** The completeness of [LJT] w.r.t. Kripke semantics follows
   from the Universal Completeness.

   The universal Kripke model consists of contexts as worlds,
   the sub-context relation, and the provability for atomic
   sentences, and term as the constant domain. *)

(** ** The Universal Model *)

(** Monotonicity of forcing relation *)

Lemma weakening_atom : forall Ga Ga',
  sub_ctx Ga Ga' ->
  forall (P : predicate * term),
    Ga |- Atom P ->
    Ga' |- Atom P.
Proof.
  intros Ga Ga' h1 P h2; exact (weakening h1 h2).
Qed.

(** Reflexivity of sub-context relation. *)

Lemma sub_ctx_refl : forall Ga, sub_ctx Ga Ga.
Proof.
  apply sub_refl.
Qed.

(** Transitivity of sub-context relation. *)

Lemma sub_ctx_trans : forall Ga Ga0 Ga1, 
  sub_ctx Ga Ga0 -> sub_ctx Ga0 Ga1 -> sub_ctx Ga Ga1.
Proof.
  unfold sublist; intros Ga Ga0 Ga1 H H0 a H1.
  apply H0; apply H; exact H1.
Qed.

(** Universal model's domain is the set of well-formed terms. *)

Canonical Structure Universal := @Build_Kripke
  context
  sub_ctx
  sub_ctx_refl
  sub_ctx_trans
  term
  (@Cst nil)
  (@App nil)
  (fun Ga P => Ga |- Atom P)
  weakening_atom.

(** In the universal model, substitution plays the role of
   semantic interpretation. *)

Lemma psem_tsubst : forall m (t:pterm m) (rho eta:assoc term),
  sub_name (op_t t) (dom rho) -> 
  psem Universal t rho eta = [[nil ! rho, eta]] t.
Proof.
  induction t; simpl; intros; auto.
    case_assoc; auto.

    remember (rho ^^ n) as r1.
    destruct r1; auto.
    assert (IN_name n (dom rho)) as H0.
      auto with datatypes.
    destruct (IN_dom_sig n rho H0).
    congruence.

    generalize (sub_appL _ _ H); intro;
        generalize (sub_appR _ _ H); intro;
          f_equal; auto.
Qed.

(** ** Universal Completeness *)

Theorem univ_cpltness : forall m (A:pfml m) (Ga:context) (rho eta:assoc term),
  sub_name (op_f A) (dom rho) ->
  Ga ||- A [rho, eta] ->
  Ga |- [nil ! rho, eta] A

with univ_cpltness_stoup : forall m (A:pfml m) (Ga:context) (rho eta:assoc term),
    sub_name (op_f A) (dom rho) ->
    (forall C Ga', sub_ctx Ga Ga' -> Ga' ;; [nil ! rho, eta] A |- C -> Ga' |- C) ->
    Ga ||- A [rho, eta].
Proof.
  destruct A as [ [p t] | | ]; simpl; intros.

  rewrite <- psem_tsubst; assumption.

  generalize (sub_appL _ _ H); intro.
  generalize (sub_appR _ _ H); intro; clear H.
  apply ProofImplyR.
  apply univ_cpltness; auto.
  apply X; auto using sub_cons_2.
  apply univ_cpltness_stoup; intros; auto.
  apply ProofCont with (A:= [nil ! rho, eta] A1); auto using IN_eq.

  set (a:= new (op_c (Forall x ([x :: nil ! rho, eta]A) :: Ga))).
  apply ProofForallR with (a:=a); intros.
  apply new_fresh.
  rewrite substitution; apply univ_cpltness; auto using sub_ctx_refl.

  destruct A as [ [p t] | |]; simpl; intros.
  
  apply H0; auto using sub_ctx_refl.
  rewrite psem_tsubst; auto using ProofAxiom.

  generalize (sub_appL _ _ H); intro.
  generalize (sub_appR _ _ H); intro; clear H.
  apply univ_cpltness_stoup; intros; auto.
  apply H0; eauto using sub_ctx_trans.
  apply ProofImplyL; eauto using univ_cpltness, mono_force.
  
  apply univ_cpltness_stoup; intros; auto.
  apply H0; eauto using sub_ctx_trans.
  apply ProofForallL with (u:=d).
  rewrite substitution; auto.
Qed.

(** ** Completeness for closed formulae *)

(** Closed formulas are formulas with no parameters. *)

Notation cl_a := (fun (A:formula) => sub_name (op_f A) nil).

Notation cl_c :=
  (fun (Ga:context) => forall A, IN_ctx A Ga -> sub_name (op_f A) nil).

Theorem univ_cpltness' : forall (A:formula)(Ga:context),
  cl_a A ->
  Ga ||- A [nil, nil] ->
  Ga |- A.
Proof.
  intros.
  rewrite freloc_id with (h:=sub_ov_fml A).
  rewrite <- pfml_subst with (eta:=nil).
  apply univ_cpltness; auto.
Qed.

Theorem univ_cpltness_stoup' : forall (A:formula)(Ga:context),
  cl_a A ->
  (forall C Ga', sub_ctx Ga Ga' -> Ga' ;; A |- C -> Ga' |- C) ->
  Ga ||- A [nil, nil].
Proof. 
  intros.
  rewrite freloc_id with (h:=sub_ov_fml A) in H0.
  rewrite <- pfml_subst with (eta:=nil) in H0.
  apply univ_cpltness_stoup; auto.
Qed.

Lemma univ_cpltness_axiom : forall (Ga:context)(A:formula),
  cl_a A ->
  IN_ctx A Ga ->
  Ga ||- A [nil, nil].
Proof.
  induction Ga as [| B]; simpl; intros A H H0;
    [ elim H0
      |  destruct (fml_dec A B); subst;
        [ apply univ_cpltness_stoup'; intros;
          eauto using ProofCont, IN_eq
          | apply mono_force with Ga; 
            [ change (sub_ctx Ga (B::Ga)); apply sub_cons_2
              | apply IHGa; assumption
            ]
        ]
    ].
Qed.  

Lemma universal_force_context : forall (Ga: worlds Universal) De,
  cl_c De ->
  sub_ctx De Ga -> 
  list_forall (fun B => Ga ||- B [nil, nil]) De.
Proof.
  induction De as [|A]; intros; constructor;
    [ apply univ_cpltness_axiom; auto using IN_eq
      | apply IHDe; eauto using sub_cons_1;
        intros; auto using IN_cons
    ].
Qed.

Theorem completeness : forall Ga A,
  cl_a A ->
  cl_c Ga ->
  (forall K (w: worlds K),
    (list_forall (fun B => w ||- B [nil, nil]) Ga) -> w ||- A [nil, nil]) ->
  Ga |- A.
Proof.
  intros; apply univ_cpltness';
    [ assumption
      | apply X; apply universal_force_context with (Ga:=Ga)(De:=Ga); auto
    ].
Qed.

(** ** Cut-admissibility *)

Definition cut_ad Ga A B (H : cl_c Ga) (H0 : cl_a B) : 
  Ga ;; A |- B -> Ga |- A -> Ga |- B :=
  fun p1 p2 =>
    completeness H0 H 
    (fun K w H1 =>
      soundness_stoup p1 K w nil H1 (soundness p2 K w nil H1)).
