(** * Kripke Semantics *)

Set Implicit Arguments.

Require Import decidable_IN.
Require Import associations.
Require Import sublist.
Require Import language_syntax.
Require Import substitution.
Require Import variable_sets.
Require Import fresh_parameters.
Require Import relocation.  

(** ** Kripke Model *)

(** We use the conventional Kripke model adopted by Troelstra and van Dalen
   for the semantics of LJT.

   - A Kripke model is a tuple of a partially-ordered set of _worlds_, a domain,
     interpretations of constants and function symbols into the domain,
     and a binary relation between worlds and atomic sentences. *)

Record Kripke : Type := {
  worlds: Type;
  wle : worlds -> worlds -> Type;
  wle_refl : forall w, wle w w;
  wle_trans : forall w w' w'', wle w w' -> wle w' w'' -> wle w w'';
  domain : Type;
  csts : name -> domain;
  funs : function -> domain -> domain -> domain;
  atoms : worlds -> predicate * domain -> Type;
  atoms_mon : forall w w', wle w w' -> forall P, atoms w P -> atoms w' P
  }.

Notation "w <= w'" := (wle _ w w') (at level 70).

(** ** Semantics *)

(** Interpretation of a pseudo-term in a Kripke model
   using associations for parameters and variables.

   - The interpretation of pseudo-terms is made total
     by ignoring insignificant parameters and variables.

   - The being-in-a-list part of a term is simply
     ignored during the interpretation.

   - The trace relocation has no impact on the Kripke semantics. *)

Fixpoint psem K m (t:pterm m)(rho:assoc (domain K))(eta:assoc (domain K)){struct t} :
  domain K :=
  match t with
    | Bvar x _ => match fun_assoc eta x with
                    | Some u => u
                    | None => csts _ 0 (* fixed value *)
                  end
    | Fvar a   => match fun_assoc rho a with
                    | Some u => u
                    | None => csts _ 0 (* fixed value *)
                  end
    | Cst c => csts _ c
    | App f t1 t2 => funs _ f (psem _ t1 rho eta) (psem _ t2 rho eta)
  end.

(** Forcing: The Kripke semantics is extended to all formulae *)

Reserved Notation "w ||- A [ rho , eta ]" (at level 80).

Fixpoint force K (w: worlds K) m (A : pfml m) rho eta {struct A} : Type :=
  match A with
    | Atom (p,t) => atoms K w (p, psem K t rho eta)
    | B --> C    => forall w', w <= w' -> 
                    w' ||- B [rho, eta] -> w' ||- C [rho, eta]
    | Forall y B => forall w', w <= w' ->
                    forall (d :domain K), w' ||- B [rho, (y, d) :: eta]
  end
  
  where " w ||- A [ rho , eta ] " := (force _ w A  rho eta).

Notation "w ||- A [ rho , eta ]" := (force _ w A rho eta) (at level 80).

(** Forcing is monotone w.r.t. the world relation. *)

Lemma mono_force : forall K (w w':worlds K) rho eta,
  w <= w' -> 
  forall m (A:pfml m),
    w ||- A [rho, eta] -> w' ||- A [rho, eta].
Proof.
  induction A as [m [p t]| |]; simpl; intros.  
  apply atoms_mon with (w:=w); auto. 

  apply X0; eauto using wle_trans. 

  apply X0; eauto using wle_trans. 
Qed.

(** Kripke semantics is independent of relocation. *)

Lemma reloc_psem : forall m (t: pterm m) K rho eta 
  k l (h: sub_name (ov_t t) k) (h0: sub_name (ov_t t) l),
  psem K (treloc t h) rho eta = psem K (treloc t h0) rho eta.
Proof. 
  induction t; simpl; intros; f_equal; auto. 
Qed.

Fixpoint reloc_force m (A: pfml m) {struct A} :
  forall K (w: worlds K) rho eta
    k l (h: sub_name (ov_f A) k) (h0: sub_name (ov_f A) l),
    w ||- (freloc A h) [rho, eta] ->
    w ||- (freloc A h0) [rho, eta] :=
      match A in pfml _ return 
        forall K (w: worlds K) rho eta
          k l (h: sub_name (ov_f A) k) (h0: sub_name (ov_f A) l),
          w ||- (freloc A h) [rho, eta] ->
          w ||- (freloc A h0) [rho, eta]
        with
        | Atom (p, t) => fun K w rho eta k l h h0 Hatom => 
          let P := fun (d:domain K) => atoms K w (p, d) in
            eq_rect_r P Hatom (
              reloc_psem t K rho eta (IN_inv_atom _ _ h0)(IN_inv_atom _ _ h))
        | B --> C => fun K w rho eta k l h h0 Himply =>
          let h' := IN_inv_impL B C h in
          let h'' := IN_inv_impR B C h in
          let h0' := IN_inv_impL B C h0 in
          let h0'' := IN_inv_impR B C h0 in
            fun w' (h1:w <= w') (h2: w' ||- (freloc B h0') [rho, eta]) =>
              let p1 := reloc_force B K w' rho eta h0' h' h2 in
              let p2 := Himply w' h1 p1 in
                reloc_force C K w' rho eta h'' h0'' p2
        | Forall y B => fun K w rho eta k l h h0 Hall => 
          fun w' (h1:w <= w')(d:domain K) =>
            reloc_force B K w' rho ((y,d) :: eta)
            (IN_inv_all B h) (IN_inv_all B h0) (Hall w' h1 d)
      end.

(** ** Equivalence of associations *)

Definition assoc_equiv (A:Type) (m:trace) (eta eta0:assoc A) :=
  forall x, IN_name x m -> eta ^^ x = eta0 ^^ x.

Hint Unfold assoc_equiv.

Lemma assoc_equiv_comm : forall (A:Type) (m:trace) (eta eta0:assoc A),
  assoc_equiv m eta eta0 ->
  assoc_equiv m eta0 eta.
Proof.
  firstorder using assoc_equiv.
Qed.

Lemma assoc_equiv_cons : forall A m (eta eta0 : assoc A),
  assoc_equiv m eta eta0 ->
  forall x d,
    assoc_equiv (x :: m) ((x, d) :: eta) ((x, d) :: eta0).
Proof.
  unfold assoc_equiv; intros; simpl; case_var; auto.
  eauto using neq_IN_cons.
Qed.

Hint Resolve assoc_equiv_comm assoc_equiv_cons.

(** A specified tactic for handling [assoc_equiv]. *)

Ltac solve_assoc :=
  let z := fresh "z" in
    let Hz := fresh "Hz" in
      intros z Hz; simpl in *; repeat case_var; intuition.

(** Semantics for well-defined terms/formulas is fixed. *)

Lemma psem_assoc_equiv : forall K m rho eta eta0 (t:pterm m),
  assoc_equiv m eta eta0 -> 
  psem K t rho eta = psem K t rho eta0.
Proof.
  induction t; simpl; intros;
    [ rewrite H; auto
      | reflexivity
      | reflexivity
      | f_equal; auto
    ].
Qed.

Lemma psem_term_assoc_indep : forall K rho eta eta' (t:term),
  psem K t rho eta = psem K t rho eta'.
Proof.
  intros; apply psem_assoc_equiv; firstorder.
Qed.  

Lemma force_assoc_equiv : forall m (A:pfml m) K (w:worlds K) rho eta eta0,
  assoc_equiv m eta eta0 -> 
  w ||- A [rho, eta] ->
  w ||- A [rho, eta0].
Proof.
  induction A as [m [p t] | |]; simpl; intros;
    [ rewrite psem_assoc_equiv with (eta0:=eta0) in X; auto
      | eauto
      | apply IHA with (eta := (x,d)::eta); auto
    ].
Qed.

Lemma force_formula_assoc_indep : forall K (w:worlds K) rho eta eta' (A:formula),
  w ||- A [rho, eta] ->
  w ||- A [rho, eta'].
Proof.
  intros; apply force_assoc_equiv with eta; firstorder.
Qed.

(** Fresh parameters have no impact on semantics. *)

Lemma psem_fresh_parameter : forall m (t:pterm m) K a d rho eta,
  a # op_t t ->
  psem K t rho eta =  psem K t ((a,d)::rho) eta.
Proof.
  induction t; simpl; intros;
    [ reflexivity
      | repeat case_var; intuition congruence
      | reflexivity
      | destruct_notin; f_equal; auto
    ].
Qed.

Lemma force_fresh_parameterL : forall m (A:pfml m) K (w:worlds K) a d rho eta,
  w ||- A [rho, eta] ->
  a # op_f A ->
  w ||- A [(a,d) :: rho, eta]

  with force_fresh_parameterR : forall m (A:pfml m) K (w:worlds K) a d rho eta,
    w ||- A [(a,d) :: rho, eta] ->
    a # op_f A ->
    w ||- A [rho, eta].
Proof.
  induction A as [ m [p t]| | ]; simpl; intros;
    [ rewrite <- psem_fresh_parameter; auto
      | destruct_notin; eauto
      | eauto
    ].  

  induction A as [ m [p t]| | ]; simpl; intros;
    [ rewrite <- psem_fresh_parameter in *; auto
      | destruct_notin; eauto
      | eauto
    ].
Qed.

(** ** Forcing and substitution *)

Theorem psem_tsubst_IN : forall m (t:pterm m) l x u K rho eta,
  sub_name (dom eta) l ->
  IN_name x l ->
  psem K ([[l ! nil, (x,u)::nil]] t) rho eta = psem K t rho eta.
Proof.
  induction t; simpl; intros;
    [ destruct (x <-? l); try case_var; auto; simpl; intuition (rewrite notIN_dom_None); auto
      | reflexivity
      | reflexivity
      | f_equal; auto ].
Qed.

Theorem force_subst_IN : forall m (A:pfml m) l x u K (w:worlds K) rho eta,
  sub_name (dom eta) l ->
  IN_name x l ->
  w ||- ([l ! nil, (x,u)::nil] A) [rho, eta] ->
  w ||- A [rho, eta]

  with subst_IN_force : forall m (A:pfml m) l x u K (w:worlds K) rho eta,
    sub_name (dom eta) l ->
    IN_name x l ->
    w ||- A [rho, eta] ->
    w ||- ([l ! nil, (x,u)::nil] A) [rho, eta].
Proof.
  induction A as [m [p t] | m A1 IHA1 A2 IHA2 | m y A]; simpl; intros.
  rewrite <- psem_tsubst_IN with (l:=l) (x:=x) (u:=u); assumption.

  clear IHA1 IHA2.
  apply force_subst_IN with (l:=l)(x:=x)(u:=u);[ assumption | assumption | ].
  apply X; try assumption. 
  apply subst_IN_force; assumption.

  clear IHA.
  apply force_subst_IN with (l:=y::l)(x:=x)(u:=u);
    [ simpl; apply sub_cons_3; exact H
      | apply IN_cons; exact H0
      | apply X; assumption ].

  induction A as [m [p t] | m A1 IHA1 A2 IHA2 | m y A]; simpl; intros.
  rewrite psem_tsubst_IN with (x:=x)(u:=u); assumption.

  clear IHA1 IHA2.
  apply subst_IN_force; try assumption.
  apply X; try assumption. 
  apply force_subst_IN with (l:=l)(x:=x)(u:=u); assumption.

  clear IHA.
  apply subst_IN_force;
    [ simpl; apply sub_cons_3; exact H
      | apply IN_cons; exact H0
      | apply X; assumption ].
Qed.

Lemma psem_tsubst : forall m (t:pterm m) x l (u:term) K rho eta,
  sub_name (dom eta) l ->
  psem K t rho ((x, psem K u rho eta) :: eta) =
  psem K ([[(rm_name x l) ! nil, (x,u):: nil]] t) rho eta.
Proof.
  induction t; simpl; intros;
    [ case_var; case_IN;
      [ elim notIN_rm_1 with name eqdec x0 l; exact i0
        | pattern u at 1; rewrite (treloc_id u (sub_ov_term u)); apply reloc_psem
        | auto
        | rewrite notIN_dom_None; auto;
          contradict n0; unfold sublist in *; apply IN_rm_2; auto
      ]          
      | reflexivity
      | reflexivity
      | f_equal; auto ].
Qed.

Theorem force_subst : forall m (A:pfml m) x l (u:term) K (w:worlds K) rho eta,
  sub_name (dom eta) l ->
  w ||- A [rho, (x, psem K u rho eta) :: eta] ->
  w ||- ([(rm_name x l) ! nil, (x,u)::nil] A) [rho, eta]

with subst_force : forall m (A:pfml m) x l (u:term) K (w:worlds K) rho eta, 
  sub_name (dom eta) l ->
  w ||- ([(rm_name x l) ! nil, (x,u)::nil] A) [rho, eta] ->
  w ||- A [rho, (x, psem K u rho eta) :: eta].
Proof.
  destruct A as [ [p t] | A1 A2 | y A ]; simpl; intros;
    [ rewrite <- psem_tsubst; [exact X | exact H]
      | apply force_subst; auto; apply X; auto; apply subst_force with (l:=l); auto
      | destruct (x == y) as [xyeq | xyneq];
        [ apply subst_IN_force; 
          [ simpl; apply sub_rm_4; assumption
            | rewrite xyeq; apply IN_eq
            | apply force_assoc_equiv with
              ((y, d) :: (x, psem K u rho eta) :: eta); auto; solve_assoc 
          ]
          | rewrite rm_cons;
            [ apply force_subst; 
              [ simpl; apply sub_cons_3; exact H
                | rewrite psem_assoc_equiv with (eta0:= eta);
                  [ apply force_assoc_equiv with
                    ((y, d) :: (x, psem K u rho eta) :: eta); auto; solve_assoc
                    | solve_assoc
                  ]
              ]
              | assumption ]
        ]
    ].

  destruct A as [ [p t] | A1 A2 | y A ]; simpl; intros;
    [ rewrite <- psem_tsubst in X; [exact X | exact H]
      | apply subst_force with (l:=l); auto; apply X; auto; apply force_subst; auto
      | destruct (eqdec x y) as [xyeq | xyneq];
        [ apply force_assoc_equiv with ((y,d) :: eta); 
          [ solve_assoc
            | apply force_subst_IN with (l:= y::rm_name x l)(x:=x)(u:=u);
              [ simpl; apply sub_rm_4; assumption
                | rewrite xyeq; apply IN_eq
                | apply X; exact X0
              ]
          ]
          | rewrite rm_cons in X;
            [ apply force_assoc_equiv with ((x, psem K u rho eta) :: (y,d) :: eta);
              [ solve_assoc
                | rewrite psem_assoc_equiv with (eta0:= (y,d)::eta);
                  [ apply subst_force with (l:= y::l);
                    [ simpl; apply sub_cons_3; exact H
                      | apply X; exact X0
                    ]
                    | solve_assoc
                  ]
              ]
              | assumption
            ]
        ]
    ].
Qed.





