(** Renaming of Parameters *)

Set Implicit Arguments.

Require Import decidable_IN.
Require Import associations.
Require Import sublist.
Require Import language_syntax.
Require Import fresh_parameters.
Require Import variable_sets.
Require Import substitution.
Require Import relocation.

(** [McKinna & Pollak] and [Leroy] used [simultaneous] parameter swapping
   in order to show the equivalence between different styles of quantification.

   We will use simultaneous [renaming] of parameters.
   Note that [renaming] is easier to deal with. *)

(** [renaming] is an arbitrary funtion from [Fvar] to [Fvar]. *)

Notation renaming := (assoc name).

Fixpoint rename_t (eta : renaming) m (t : pterm m) : pterm m := 
  match t with
    | Bvar y h    => Bvar _ y h
    | Fvar a      => match eta ^^ a with
                       | Some d => Fvar _ d
                       | None   => Fvar _ a
                     end
    | Cst c       => Cst _ c
    | App f t1 t2 => App f (rename_t eta t1)  (rename_t eta t2)
  end.

Fixpoint rename_f (eta : renaming) m (A : pfml m) : pfml m := 
  match A with
    | Atom (p,t) => Atom (p, rename_t eta t )
    | B --> C    => (rename_f eta B) --> (rename_f eta C)
    | Forall y B => Forall y (rename_f eta B)
  end.

Fixpoint rename_c (eta : renaming) (Ga : context) : context := 
  match Ga with
    | nil => nil
    | A :: Ga0 => rename_f eta A :: rename_c eta Ga0
  end.

(** Renaming associations *)

Fixpoint rename_a (eta : renaming) (rho : assoc term) :=
  match rho with
    | nil           => nil
    | (x,t) :: rho0 => (x, rename_t eta t) :: rename_a eta rho0
  end.

(** No renaming means no change  *)

Lemma rename_t_nil : forall m (t:pterm m),
  rename_t nil t = t.
Proof.
  induction t; simpl; f_equal; auto.
Qed.

Lemma rename_f_nil : forall m (A:pfml m),
  rename_f nil A = A.
Proof.
  induction A as [m [p t]| |]; simpl; f_equal; auto.
  f_equal; auto using rename_t_nil.
Qed.

Lemma rename_c_nil : forall (Ga:context),
  rename_c nil Ga = Ga.
Proof.
  induction Ga; simpl; f_equal; auto using rename_f_nil.
Qed.

(** Compatibility of renaming with predicates and operators. *)

Lemma rename_IN_ctx : forall A Ga eta,
  IN_ctx A Ga ->
  IN_ctx (rename_f eta A) (rename_c eta Ga).
Proof.
  induction Ga; simpl; intros;
  repeat case_lang; auto; congruence.
Qed.

Lemma IN_dom_sig_rename : forall (x: name) (rho: assoc term)(eta:renaming) t,
  rho ^^ x = Some t ->
  (rename_a eta rho) ^^ x = Some (rename_t eta t).
Proof.
  induction rho as [ | [y b] ]; simpl; intros;
    [ discriminate
      | case_var; inversion H; subst; auto
    ].
Qed.

Lemma treloc_rename_t : forall m l (t:pterm m) eta
  (H: sub_name (ov_t t) l) (H0 : sub_name (ov_t (rename_t eta t)) l),
  rename_t eta (treloc t H) = treloc (rename_t eta t) H0.
Proof.
  induction t; simpl; intros;
    [ f_equal; apply IN_proof_uniq
      | case_assoc; auto
      | reflexivity
      | f_equal; auto
    ].
Qed.

(** Renaming and substitution *)

Lemma rename_tsubst : forall m (t:pterm m) (rho:assoc term)(eta:renaming) l,
  sub_name m (l ++ dom rho) ->
  rename_t eta ([[l ! nil, rho]] t) =
  [[l ! nil, rename_a eta rho]] (rename_t eta t).
Proof.
  induction t; simpl; intros;
    [ destruct (IN_app_or (H x i));
      [ destruct (x <-? l); simpl; auto; contradiction
        | destruct (x <-? l); simpl; auto;
          destruct (IN_dom_sig x rho H0); rewrite e;
            rewrite (IN_dom_sig_rename x rho eta e); apply treloc_rename_t
      ]
      | case_assoc; auto
      | reflexivity
      | f_equal; auto
    ].
Qed.

Lemma rename_subst : forall m (A:pfml m) (rho:assoc term)(eta:renaming) l,
  sub_name m (l ++ dom rho) ->
  rename_f eta ([l ! nil, rho] A) =
  [l ! nil, rename_a eta rho] (rename_f eta A).
Proof.
  induction A as [m [p t] | |]; simpl; intros;
    [ do 2 f_equal; apply rename_tsubst; auto
      | f_equal; auto
      | f_equal; apply IHA; unfold sublist; intros; simpl in *; case_var; auto
    ].
Qed.

(** ** Renaming and freshness *)

Lemma rename_t_fresh : forall m (t:pterm m) eta a a0,
  a # op_t t ->
  rename_t ((a,a0) :: eta) t = rename_t eta t.
Proof.
  induction t; simpl; intros;
    [ reflexivity
      | repeat case_var; try congruence; elim H; auto
      | reflexivity
      | destruct_notin; f_equal; auto
    ].
Qed.

Lemma rename_f_fresh : forall m (A:pfml m) eta a a0,
  a # op_f A ->
  rename_f ((a, a0) :: eta) A = rename_f eta A.
Proof.
  induction A as [m [p t] | | ]; simpl; intros;
    [ do 2 f_equal; apply rename_t_fresh
      | destruct_notin; f_equal
      | f_equal
    ]; auto.
Qed.

Lemma rename_c_fresh : forall Ga eta a a0,
  a # op_c Ga ->
  rename_c ((a, a0) :: eta) Ga = rename_c eta Ga.
Proof.
  induction Ga as [ | A Ga0]; simpl; intros; auto.
  destruct_notin; f_equal; auto using rename_f_fresh.
Qed.

(** Function values and (co-)domain *)

Fixpoint image (eta : renaming) : list name :=
  match eta with
    | nil => nil
    | (_, a0) :: eta0 => a0 :: image eta0
  end.

Definition fun_value (eta : renaming) (a:name) : name :=
  match eta ^^ a with
    | Some a0 => a0
    | None   => a
  end.

Notation " eta ** a " := (fun_value eta a) (at level 60).

Lemma notIN_image : forall (eta:renaming) a a0 a1,
  a # image eta ->
  eta ^^ a0 = Some a1 ->
  a <> a1.
Proof.
  induction eta as [ |[x] ]; simpl; intros;
  repeat case_var; intuition try congruence; eauto.
Qed.

Lemma pterm_rename_fresh : forall m (t:pterm m) eta a a0,
  a0 # op_t t ->
  a0 # dom eta ->
  a0 # image eta ->
  rename_t ((a0, eta ** a) :: eta) (rename_t ((a, a0) :: nil) t) = 
  rename_t eta t.
Proof.
  induction t; simpl; intros; destruct_notin;
    [ reflexivity
      | repeat case_var; simpl; firstorder;
        [ rewrite (notIN_dom_None _ _ H0); simpl; 
          case_var; intuition; unfold fun_value;
            destruct (eta ^^ a); auto
          | case_eq (eta ^^ n); intros;
            [ assert (n2 <> a0);
              [ apply sym_not_eq; eauto using notIN_image
                | simpl; case_var; congruence
              ]
              | simpl; case_var; congruence
            ]
        ]
      | reflexivity
      | f_equal; auto].
Qed.

Lemma pfml_rename_fresh : forall m (A:pfml m) eta a a0,
  a0 # op_f A ->
  a0 # dom eta ->
  a0 # image eta ->
  rename_f ((a0, eta ** a) :: eta) (rename_f ((a, a0) :: nil) A) = 
  rename_f eta A.
Proof.
  induction A as [m [p t]| |]; simpl; intros;
    [ do 2 f_equal; auto using pterm_rename_fresh
      | destruct_notin; f_equal; auto
      | f_equal; auto
    ].
Qed.

Lemma context_rename_fresh : forall De eta a a0,
  a0 # op_c De ->
  a0 # dom eta ->
  a0 # image eta ->
  rename_c ((a0, eta ** a) :: eta) (rename_c ((a, a0) :: nil) De) =
  rename_c eta De.
Proof.
  induction De as [ | A]; simpl; intros;
    [ reflexivity
      | destruct_notin; f_equal; auto using pfml_rename_fresh
    ].
Qed.

Lemma notIN_op_rename_t : forall m (t : pterm m) eta a a0,
  a0 # image eta ->
  a <> a0 ->
  a0 # op_t (rename_t ((a0, eta ** a) :: eta) t).
Proof.
  induction t; simpl; intros;
    [ auto
      | case_var; simpl;
        [ unfold fun_value; case_eq (eta ^^ a); intros;
          [ case_var; auto; elim (notIN_image eta a H H1); auto
            | case_var; auto
          ]
          | case_eq (eta ^^ n); simpl; intros;
            [ case_var; auto; elim (notIN_image eta n H H1); auto
              | case_var; auto
            ]
        ]
      | auto
      | solve_notin
    ].
Qed.

Lemma notIN_op_rename_f : forall m (A : pfml m) eta a a0,
  a0 # image eta ->
  a <> a0 ->
  a0 # op_f (rename_f ((a0, eta ** a) :: eta) A).
Proof.
  induction A as [m [p t] | |]; simpl; intros;
    [ auto using notIN_op_rename_t
      | solve_notin
      | auto
    ].
Qed.

Lemma notIN_op_rename_c : forall (Ga : context) eta a a0,
  a0 # image eta ->
  a <> a0 ->
  a0 # op_c (rename_c ((a0, eta ** a) :: eta) Ga).
Proof.
  induction Ga as [ | A Ga0]; simpl; intros;
    auto using notIN_app_2, notIN_op_rename_f.
Qed.

Lemma sub_ctx_fresh_fv : forall Ga Ga0 a a0,
  sub_ctx Ga Ga0 ->
  a # op_c Ga ->
  sub_ctx Ga (rename_c ((a, a0) :: nil) Ga0).
Proof.
  unfold sublist; intros.
  assert (a1 = rename_f ((a, a0):: nil) a1);
    [ rewrite rename_f_fresh;
      [ symmetry; apply rename_f_nil
        | contradict H0; exact (IN_free_sub H1 a H0)
      ]
      | rewrite H2; apply rename_IN_ctx; auto
    ].  
Qed.

