(** * Substitution *)

Set Implicit Arguments.

Require Import decidable_IN.
Require Import associations.
Require Import sublist.
Require Import language_syntax.
Require Import variable_sets.
Require Import relocation.


(** Defining substitution requires special consideration for two reasons.

   - First, the relationship between syntax and semantics such as soundness
     and completeness should be realized in a natural way.

     For this purpose, we find that simultaneous substitution provides
     a convenient way in establishing the relationship.

   - Second, because of the trace part, it is not clear to which class
     the resulting expression of a substitution should belong.

     Thus we decided to state clearly the type of the resulting expression
     in the definition of substitution.


   - Notice the special role of the [Bvar] case because it needs relocation.

   - No alpha-conversion is used. *)

Fixpoint tsubst (rho eta:assoc term) l m (t: pterm m) {struct t} : pterm l :=
  match t with
    | Bvar y h    => match y <-? l with
                       | left p  => @Bvar l y p
                       | right _ => match eta ^^ y with
                                      | Some u => (treloc u (term_sub u l))
                                      | None   => @Cst l 0
                                    end
                     end
    | Fvar a      => match rho ^^ a with
                       | Some u => (treloc u (term_sub u l))
                       | None   => @Fvar l a
                     end
    | Cst c       => @Cst l c
    | App f t1 t2 => App f (tsubst rho eta l t1)  (tsubst rho eta l t2)
  end.

Notation "[[ l ! rho , eta ]] t" := (tsubst rho eta l t) (at level 55, rho , eta at next level).

Fixpoint subst (rho eta: assoc term) l m (A:pfml m) {struct A} : pfml l :=
  match A with
    | Atom (p,t) => Atom (p, [[l ! rho, eta]] t )
    | B --> C    => (subst rho eta l B) --> (subst rho eta l C)
    | Forall y B => Forall y (subst rho eta (y::l) B)
  end.

Notation "[ l ! rho , eta ] A" := (subst rho eta l A) (at level 55, rho , eta at next level).

Lemma pterm_tsubst : forall m (t:pterm m) eta l (H: sub_name (ov_t t) l),
  [[l ! nil, eta]] t = treloc t H.  
Proof.
  induction t; simpl; intros;
    [ destruct (x <-? l) as [|Hnotin]; [ f_equal | elim Hnotin];
      auto with datatypes
      | reflexivity
      | reflexivity
      | f_equal; auto ].
Qed.

(** For well-defined term, local variable substitution has no effect. *)

Lemma term_tsubst : forall (t:term) (eta:assoc term),
  [[nil ! nil, eta]] t = t.
Proof.
  intros;
  assert (sub_name (ov_t t) nil) as Hsub ;
    [ rewrite term_ov_nil_1; apply sub_refl
      | transitivity (treloc t Hsub); auto using pterm_tsubst
    ].        
Qed.

Lemma pfml_subst : forall m (A:pfml m) eta l (H: sub_name (ov_f A) l),
  [l ! nil, eta] A = freloc A H.  
Proof.
  induction A as [m [p t] | |]; simpl; intros; f_equal; auto.
  f_equal; auto using pterm_tsubst.
Qed.

(** For well-defined formulae, local variable substitution has no effect. *)

Lemma fml_subst : forall (A:formula) (eta:assoc term),
  [nil ! nil, eta] A = A.
Proof.
  intros; assert (sub_name (ov_f A) nil) as Hsub ;
    [ rewrite fml_ov_nil_1; apply sub_refl
      | transitivity (freloc A Hsub); auto using pfml_subst
    ].
Qed.

(** Relocation has no effect on substitution. *)

Lemma tsubst_treloc_1 : forall m (t:pterm m) l l' rho eta (h: sub_name (ov_t t) l),
  [[l' ! rho, eta]] (treloc t h) = [[l' ! rho, eta]] t.
Proof.
  induction t; simpl; intros; f_equal; auto. 
Qed.

Lemma tsubst_reloc_2 : forall m (t:pterm m) l l' eta
  (h1: sub_name (ov_t t) l) (h2: sub_name (ov_t t) l'),
  [[l' ! nil, eta]] (treloc t h1) = treloc t h2.
Proof.
  intros; transitivity ([[l' ! nil, eta]] t);
    auto using tsubst_treloc_1, pterm_tsubst.
Qed.

(** Extending substitution with forbidden variables has no impact. *)

Lemma tsubst_IN : forall m (t:pterm m) x l v rho eta,
  IN_name x l ->
  [[l ! rho, eta]] t = [[l ! rho, (x,v)::eta]] t.
Proof.
  induction t; simpl; intros;
    [ case_IN; auto; case_var; auto; case_assoc; auto; elim n; exact H
      | reflexivity
      | reflexivity
      | f_equal; auto
    ].
Qed.

Lemma subst_IN : forall m (A:pfml m) x l u rho eta,
  IN_name x l ->
  [l ! rho, eta] A = [l ! rho, (x,u)::eta] A.
Proof.
  induction A as [m [p t] | | ]; simpl; intros;
    [ rewrite <- tsubst_IN; [reflexivity | exact H]
      | f_equal; auto
      | f_equal; auto using IN_cons
    ].
Qed.

(** Substitution and [LEq] *)

Lemma tsubst_LEq_1 : forall m (t:pterm m) l l' rho eta eta',
  LEq l l' ->
  [[l' ! nil, eta']] ([[l ! rho, eta]] t) = [[l' ! rho, eta]] t.
Proof.
  induction t; simpl; intros;
    [ destruct (x <-? l); simpl; destruct (x <-? l'); try reflexivity;
      [ elim n; elim H; intros; apply H0; exact i0
        | elim n; elim H; intros; apply H1; exact i0
        | destruct (fun_assoc eta x); simpl; try reflexivity;
          apply tsubst_reloc_2; apply term_ov_nil_1
      ]
      | case_assoc; simpl;
        [ rewrite tsubst_treloc_1; apply pterm_tsubst
          | auto
        ]
      | reflexivity
      | f_equal; auto
    ].
Qed.
  
Lemma subst_LEq_1 : forall m (A:pfml m) l l' rho eta eta',
  LEq l l' ->
  [l' ! nil, eta'] ([l ! rho, eta] A) = [l' ! rho, eta] A.
Proof.
  induction A as [m [p t] | | ]; simpl; intros l l' rho eta eta' H; inversion H;
    [ do 2 f_equal; apply tsubst_LEq_1; auto
      | f_equal; auto
      | f_equal; auto using LEq_cons
    ].
Qed.

Lemma tsubst_LEq_2 : forall m (t:pterm m) l l' k rho eta eta',
  LEq l l' ->
  [[k ! nil, eta']] ([[l ! rho, eta]] t) = [[k ! nil, eta']] ([[l' ! rho, eta]] t).
Proof.
  induction t; simpl; intros;
    [ inversion H; repeat (case_IN); auto;
      [ elim n; auto
        | elim n; auto
        | case_assoc; auto;
          transitivity (treloc p (term_sub p k));
            eauto using tsubst_reloc_2
      ]
      | case_assoc;
        [ do 2 rewrite tsubst_treloc_1; reflexivity
          | reflexivity
        ]
      | reflexivity
      | f_equal; auto
    ].
Qed.

Lemma subst_LEq_2 : forall m (A:pfml m) l l' k rho eta eta',
  LEq l l' ->
  [k ! nil, eta'] ([l ! rho, eta] A) = [k ! nil, eta'] ([l' ! rho, eta] A).
Proof.
  induction A as [m [p t] | | ]; simpl; intros;
    [ do 2 f_equal; auto using tsubst_LEq_2
      | f_equal
      | f_equal
    ]; auto using LEq_cons.
Qed.

(** ** Substitutions Lemma *)

Lemma t_substitution : forall m (t:pterm m) y l (rho eta:assoc term) (u:term),
  [[l ! nil, (y,u)::nil]] ([[(y :: l) ! rho, eta]] t) =
  [[l ! rho, (y, u) :: eta]] t.
Proof.
  induction t; intros;
    [ unfold tsubst at 2; destruct (x == y); subst;
      [ destruct (y <-? (y::l));
        [ simpl; case_var; congruence
          | subst; generalize (IN_eq eqdec y l); intro; congruence
        ]
        | destruct (x <-? (y::l)); simpl; case_var; try congruence;
          [ destruct (x <-? l); auto; contradict n1; eauto with datatypes
            | generalize (notIN_cons_2 eqdec x y l); intros; destruct (eta ^^ x); simpl;
              [ destruct (x <-? l); intuition try congruence;
                rewrite pterm_tsubst
                with (t:= treloc p (term_sub p (y ::l))) (H:= treloc_ov_2 p _ _);
                  symmetry; apply treloc_idempotent
                | destruct (x <-? l); intuition congruence
              ]
          ]
      ]
      | simpl; case_assoc; simpl;
        [ assert (sub_name (ov_t (treloc p (term_sub p (y :: l)))) l);
          [ rewrite treloc_ov_1; firstorder
            | rewrite pterm_tsubst with(H:=H); 
              rewrite <- treloc_idempotent with (h3:= term_sub p l); reflexivity
          ]
          | reflexivity
        ]
      | reflexivity
      | simpl; f_equal; auto
    ].
Qed.

Lemma substitution : forall m (A : pfml m) y l (rho eta:assoc term) (u:term),
  [l ! nil, (y,u) :: nil] ([y :: l ! rho, eta] A) = [l ! rho, (y, u) :: eta] A.
Proof.
  induction A as [m [p t] | | ]; simpl; intros;
    [ do 2 f_equal; apply t_substitution
      | f_equal; auto
      | f_equal; destruct (eqdec x y) as [xy | nxy];
        [ subst y; rewrite subst_LEq_1; auto using LEq_xx, subst_IN, IN_eq
          | rewrite subst_LEq_2 with (l':= y::x::l); auto using LEq_xy
        ]
    ].
Qed.

