(** * Language Syntax *)

Set Implicit Arguments.

Require Import decidable_IN.

(** The language syntax for the first-order logic is represented.

   - Implication and universal quantification are the only logical symbols.

   - We assume there are two decidable sets of infinitely many "unary" predicates
     and "binary" functions.

   - This is general enough because predicate and function symbols of different
     arities can be encoded. *)

Parameter predicate : Set.      (* unary predicates *)
Parameter function : Set.       (* binary functions *)

Axiom pred_dec : decidable predicate.
Axiom fun_dec : decidable function.

(* ####################################################### *)
(** *** Representation with locally traced names *)

(** We suggest a style called "representation with locally traced names".

   - Our style is closely related to the McKinna-Pollack's _locally-named_ style.

   - That is, two sorts of named variables are used:
     -- (locally bound) _variables_
     -- _parameters_ (also called _free_ variables)

   - A new feature of is that the sets of terms and formulae depend on
     a list of variables, called a _trace_, that might be used
     during the term or formula construction.

   - That is, variables occurring in an expression are controlled by
     the type of that expression.

   - An important consequence is that we can talk about _well-formed_ terms
     and formulae without defining an extra syntax.
     (Well-formed expressions are those with an empty trace.) *)

(* ####################################################### *)
(** *** Constants as parameters *)

(** Syntactically, We don't use parameters (free variables).

   - We let constants play the role of constants.

   - An important consequence at the syntactic level is that substitution
     is needed only for variables.

   - The language itself needs to contain infinitely many constants.

   - Note that every language can be conservatively extended to a language
     with infinitely many constants.

   - Even if we formally introduce parameters, the substitution for parameters
     do not play any role

   - For the formalization with parameters, see the Coq files, called [with_FreeVar_*.v]. *)

(* ####################################################### *)
(** ** Pseudo-terms *)

(** Any decidable, denumerable set can be used for the representation of variables
   and constants.
   Here we use [name = nat], the set of natural numbers, to make the formlization
   as simple as possible. 

   Given a _trace_ [m], [pterm m] is the set of pseudo-terms with free occurrence
   of variables from the trace [m]. *)

Notation trace := (list name).

Inductive pterm (m:trace) : Set := 
| Bvar : forall (x: name), IN_name x m -> pterm m
| Cst  : name -> pterm m
| App  : function -> pterm m -> pterm m -> pterm m.

(** ** Pseudo-formulas *)

(** Given a trace [m], [pfml m] is the set of pseudo-formulae with free occurrence of
   variables from the trace [m].

   - Atomic pseudo-formulas are defined first. *)

Definition atom m := (predicate * pterm m)%type.

Inductive pfml (m:trace) : Set :=
| Atom   : atom m -> pfml m
| Imply  : pfml m -> pfml m -> pfml m
| Forall : forall (x:name), pfml (x::m) -> pfml m.

Implicit Arguments Forall [m].

Notation "A --> B" := (@Imply _ A B) (at level 69, right associativity).

(** ** Well-formed terms and formulas *)

(** Well-formed expressions are expressions without any free occurrence of variables. *)

Notation term := (pterm nil).

Notation formula := (pfml nil).

(** ** Contexts *)

Notation context := (list formula).

(** ** Occurrence of constances ([oc]): *)

(** This is a necessary step to talk about fresh constants for a context. *)

Fixpoint oc_t m (t:pterm m){struct t} : list name :=
  match t with
  | Bvar _ _    => nil
  | Cst c       => c :: nil
  | App _ t0 t1 => (oc_t t0) ++ (oc_t t1)
  end.

Fixpoint oc_f m (A:pfml m){struct A} : list name :=
  match A with
  | Atom (p, t) => oc_t t
  | Imply B C   => oc_f B ++ oc_f C
  | Forall x A  => @oc_f (x::m) A
  end.

Fixpoint oc_c (Ga:context){struct Ga} : list name :=
  match Ga with
  | nil      => nil
  | A :: Ga' => oc_f A ++ oc_c Ga'
  end.

(** ** Free occurrence of variables ([ov]): *)

Fixpoint ov_t m (t:pterm m){struct t} : trace :=
  match t with
  | Bvar x _    => x :: nil
  | Cst _       => nil
  | App _ t0 t1 => (ov_t t0) ++ (ov_t t1)
  end.

Fixpoint ov_f m (A:pfml m){struct A} : trace :=
  match A with
  | Atom (p, t) => ov_t t
  | Imply B C   => ov_f B ++ ov_f C
  | Forall x B  => rm_name x (@ov_f (x::m) B)
  end.

(** ** Decidability of syntactic equality *)

(** A specified tactic for uninteresting cases *)

Ltac neq_case :=
  match goal with
    | H : ?x <> ?y |- _ =>
      right; red; intro Heq; inversion Heq; congruence
  end.

(** [pfml] is decidable. *)

Lemma term_dec : forall m (t t0 : pterm m), {t = t0} + {t <> t0}.
Proof.
  induction t; induction t0; try (right; discriminate);
    [ destruct (eqdec x x0); subst;
      [ left; rewrite (IN_proof_uniq eqdec x0 m i0 i); auto
        | neq_case
      ]
      | destruct (eqdec n n0); subst;
        [ left; auto
          | neq_case
        ]
      | destruct (fun_dec f f0); subst;
        [ destruct (IHt1 t0_1); subst;
          [ destruct (IHt2 t0_2); subst;
            [ left; reflexivity
              | neq_case
            ]
            | neq_case
          ]
          | neq_case
        ]
    ].
Qed.

Lemma atom_dec : forall m (P Q:atom m), {P = Q} + {P <> Q}.
Proof.
  destruct P as [p t]; destruct Q as [q u].
  decide equality; [ apply term_dec | apply pred_dec ].
Qed.

Require Import Eqdep_dec.

Module BvarEqdec.
  Definition U := name.
  Definition eq_dec := eqdec.
End BvarEqdec.

Module EqdepBvar := DecidableEqDep BvarEqdec.

Import EqdepBvar.

Lemma fml_dec : forall m (A B:pfml m), {A = B} + {A <> B}.
Proof.
  induction A as [? P | |]; destruct B as [Q | |]; try (right; discriminate);
    [ destruct (atom_dec P Q); 
      [ subst; left; reflexivity
        | neq_case
      ]
      | destruct (IHA1 B1); destruct (IHA2 B2);
        [ subst; left; reflexivity
          | neq_case
          | neq_case
          | neq_case
        ]
      | destruct (eqdec x x0);
        [ subst x;
          destruct (IHA B) as [ | neq];
            [ left; subst; reflexivity
              | right; intro Heq; elim neq;
                inversion Heq as [HdepEq];
                  apply inj_pairT2 in HdepEq; assumption
            ]
          | neq_case
        ]
    ].
Qed.

(** A tactic for destructing the decidable equality between pseudo-expressions. *)

Ltac case_lang :=
  let destr t u := destruct (term_dec t u); [try subst t | idtac] in
    let destr A B := destruct (fml_dec A B); [try subst A | idtac] in
  match goal with
    | |- context [term_dec ?t ?u]       => destr t u
    | _ : context [term_dec ?t ?u] |- _ => destr t u
    | |- context [fml_dec ?A ?B]       => destr A B
    | _ : context [fml_dec ?A ?B] |- _ => destr A B
  end.

