(** * Variable Sets *)

Set Implicit Arguments.

Require Import decidable_IN.
Require Import sublist.
Require Import language_syntax.

(** Subcontext relation *)

Notation sub_ctx := (@sublist formula (@fml_dec nil)).

(** Being-in-a-context relation *)

Notation IN_ctx := (IN (@fml_dec nil)).

(** Relation between context and contants *)

Lemma IN_oc_sub (A:formula) (Ga:context) :
  IN_ctx A Ga ->
  sub_name (oc_f A) (oc_c Ga).
Proof.
  unfold sublist; induction Ga as [ | B]; simpl; intros; auto.
  case_lang; subst; eauto with datatypes.
Qed.

Implicit Arguments IN_oc_sub [A Ga].

Lemma sub_ctx_oc : forall (Ga De:context),
  sub_ctx Ga De -> sub_name (oc_c Ga) (oc_c De).
Proof.
  unfold sublist; intros; induction Ga as [ | A]; simpl in * ; intros; intuition.
  destruct (IN_app_or H0). 
    generalize (H A); intros; case_lang; eauto 4 with datatypes.
    apply (IN_oc_sub (H2 I)); auto.

    apply IHGa; auto.
    intros B HB; generalize (H B); case_lang; tauto.
Qed.

Hint Resolve IN_oc_sub sub_ctx_oc.

(** Relation between variables and traces

 - Traces contain free occurrences of variables in pseudo-terms and -formulae.

 - Well-defined terms or formulas have no local variables. *)

Lemma sub_ov_term : forall m (t:pterm m),
  sub_name (ov_t t) m.
Proof.
  unfold sublist; induction t; simpl; intros; try case_var; intuition.
  destruct (IN_app_or H); eauto with datatypes.
Qed.

Lemma sub_ov_fml : forall m (A:pfml m),
  sub_name (ov_f A) m.
Proof.
  unfold sublist; induction A as [ m [p t] | |]; simpl; intros.

  apply sub_ov_term with t; assumption.

  elim (IN_app_or H); auto.

  apply neq_IN_cons with x; eauto 3 with datatypes.
  apply sym_not_eq; apply IN_rm_neq with eqdec (ov_f A); auto.
Qed.

Hint Resolve sub_ov_term sub_ov_fml.

Lemma term_ov_nil_1 : forall (t:term),
  ov_t t = nil.
Proof.
  intro t; apply sub_nil with (decA := eqdec); auto.
Qed.

Hint Rewrite -> term_ov_nil_1 : bv.

Lemma term_ov_nil_2 : forall (t:term)(x:name),
  x # (ov_t t).
Proof.
intros; rewrite term_ov_nil_1; firstorder. 
Qed.

Lemma term_sub : forall (t:term) m,
  sub_name (ov_t t) m.
Proof.
intros.
exact (eq_rect_r (fun l => sublist eqdec l m) (nil_sub m) (term_ov_nil_1 t)).
Qed.

Hint Resolve term_ov_nil_1 term_ov_nil_2 term_sub.

Lemma fml_ov_nil_1 : forall (A:formula),
  ov_f A = nil.
Proof.
intro A; eapply sub_nil; eauto.
Qed.

Hint Rewrite -> fml_ov_nil_1 : bv.

Lemma fml_ov_nil_2 : forall (A:formula) m,
  sub_name (ov_f A) m.
Proof.
intros; rewrite fml_ov_nil_1; apply nil_sub.
Qed.

Lemma fml_ov_nil_3 : forall (A:formula)(x:name),
  x # (ov_f A).
Proof.
intros; rewrite fml_ov_nil_1; firstorder. 
Qed.

Hint Resolve fml_ov_nil_1 fml_ov_nil_2 fml_ov_nil_3.

(** inversion properties for variable sets:

   - These properties are used in the definition of substitution. *)

Lemma sub_ov_funL : forall (f:function) m (t0 t1: pterm m) l,
  sub_name (ov_t (App f t0 t1)) l ->
  sub_name (ov_t t0) l.
Proof.
simpl; intros; eauto 2 with datatypes.
Qed.

Lemma sub_ov_funR : forall (f:function) m (t0 t1: pterm m) l,
  sub_name (ov_t (App f t0 t1)) l ->
  sub_name (ov_t t1) l.
Proof.
simpl; intros; eauto 2 with datatypes.
Qed.

Hint Resolve sub_ov_funL sub_ov_funR.

Lemma IN_inv_atom : forall (p:predicate) m l (t: pterm m),
  sub_name (ov_f (Atom (p,t))) l ->
  sub_name (ov_t t) l.
Proof.
intros; auto. 
Qed.

Lemma IN_inv_impL : forall m l (B C: pfml m),
  sub_name (ov_f (B --> C)) l ->
  sub_name (ov_f B) l.
Proof.
  simpl; intros; eauto 2 with datatypes.
Qed.

Lemma IN_inv_impR : forall m l (B C: pfml m),
  sub_name (ov_f (B --> C)) l ->
  sub_name (ov_f C) l.
Proof.
  simpl; intros; eauto 2 with datatypes.
Qed.

Lemma IN_inv_all : forall m l x (B:pfml (x::m)),
  sub_name (ov_f (@Forall m x B)) l ->
  sub_name (@ov_f (x::m) B) (x::l).
Proof.
  simpl; intros; auto with datatypes.
Qed.

Hint Resolve IN_inv_atom IN_inv_impL IN_inv_impR IN_inv_all.

